import cx from "clsx";
import { type FunctionComponent, forwardRef } from "react";

interface WithClassName<P = object> extends FunctionComponent<P> {
  className: string;
}

type InputProps = React.DetailedHTMLProps<
  React.InputHTMLAttributes<HTMLInputElement>,
  HTMLInputElement
>;

export const classed: {
  (
    type: "input",
    className: string | string[],
    defaultProps?: Partial<InputProps>
  ): InputProps;
  <K extends keyof JSX.IntrinsicElements>(
    type: K,
    className: string | string[],
    defaultProps?: Partial<JSX.IntrinsicElements[K]>
  ): JSX.IntrinsicElements[K];
  (
    type: string,
    className: string | string[],
    defaultProps?: Record<string, unknown>
  ): WithClassName<React.ClassAttributes<any> & React.DOMAttributes<any>>;
  <P>(
    type: FunctionComponent<P>,
    className: string | string[],
    defaultProps?: Partial<P>
  ): WithClassName<P>;
  <P>(
    type: React.ComponentClass<P>,
    className: string | string[],
    defaultProps?: Partial<P>
  ): WithClassName<P>;
} = (
  Component: any,
  classNameInput: string | string[],
  defaultProps?: Record<string, unknown>
) => {
  const className = cx(classNameInput);
  const component: any = forwardRef<any, any>(({ className: cls, ...props }, ref) => (
    <Component
      {...defaultProps}
      {...props}
      ref={ref}
      className={
        typeof cls === "function"
          ? (...args: unknown[]) => cx(className, cls(...args))
          : cx(className, cls)
      }
    />
  ));
  component.className = className;
  return component;
};
